/// This file contains all functions that interact with the database, such as
/// logging in and registering users,
/// sending data to the database,
/// getting data from the database

import 'dart:developer';
import 'dart:io';
import 'dart:typed_data';
import 'package:billettapp_for_vitensenteret/resources/structs.dart';
import 'package:firebase_auth/firebase_auth.dart';
import 'package:firebase_database/firebase_database.dart';
import 'package:firebase_storage/firebase_storage.dart';

/// Log in via firebase using password, [pass], and [email].
/// Returns 0 for success, 1 for wrong password and 2 for could not log in
Future<int> firebaseLogIn (var pass, var user) async {


  //Authenticate with Firebase Authentication
  try {
    final credential = await FirebaseAuth.instance.signInWithEmailAndPassword(
        email: user,
        password: pass
    );
    if (credential != Null) {
      return 2;
    }
  } on FirebaseAuthException catch (e) {
    if (e.code == 'user-not-found') {
      log('No user found for that email.');
      return 1;
    } else if (e.code == 'wrong-password') {
      log('Wrong password provided for that user.');
      return 1;
    }
  }
  return 0;
}

/// Register new user to database using password, [pass], and email adress, [user],
/// note that this function does not perform any checks on password strength, or the validity of the email.
/// Returns 0 for success, 1 for password too weak, 2 for email already in use, 3 for could not create user
Future<int> firebaseRegister (var pass, var user) async {

  //Attempt to create a new user in firebase using Firebase Authenticate
  try {
    final credential = await FirebaseAuth.instance.createUserWithEmailAndPassword(
      email: user,
      password: pass,
    );
    if(credential != null) {
      return 3;
    }

  } on FirebaseAuthException catch (e) {
    if (e.code == 'weak-password') {
      log('The password provided is too weak.');
      return 1;
    } else if (e.code == 'email-already-in-use') {
      log('The account already exists for that email.');
      return 2;
    }
  } catch (e) {
    log(e.toString());
  }

  return 0;
}

/// When a user successfully registers all their data is sent to the database, this
/// includes their full [name], [mail] address, [image], [mobileNum] and [birthday].
/// Data is saved under a new node with [uid] as name. Admin and employee values
/// are set to false as a default.
/// Returns 0 for success, 1 for error uploading profile picture, 2 for error uploading user data
Future<int> firebaseSendRegisterInfo (var name, var mail, var image, String uid, String mobileNum, String birthday, File profilePic) async {

  //Location of user profile image
  final storageRef = FirebaseStorage.instance.ref("userPhotos/$uid");

  String profilePicString = "";

  //Upload profile picture and get link
  try {
    await storageRef.putFile(profilePic);
    profilePicString = await storageRef.getDownloadURL();
  } catch (e) {
    log("Could not upload profile image, Error $e");
    return 1;
  }

  //Set user data in Firebase
  try {
    FirebaseDatabase.instance.ref().child("/users/$uid").set({
      'name': name,
      'email': mail,
      'phone': mobileNum,
      'birthday': birthday,
      'isAdmin': false,
      'isEmployee': false,
      'image': profilePicString,
      'tickets': ""
    });
  } catch (e) {
    log("Could not set user data, Error $e");
    return 2;
  }

  FirebaseAuth.instance.currentUser!.updatePhotoURL(profilePicString);
  FirebaseAuth.instance.currentUser!.updateDisplayName(name);
  FirebaseAuth.instance.currentUser!.updateEmail(mail);

  return 0;

}


///Gets information about all tickets a user owns from firebase
///Returns 0 if there are no issues, 1 if the user does not exist
Future<int> firebaseGetTickets() async {

  //Check if user exist in database
  bool userExists = false;
  var tickets;

  DatabaseReference ref = FirebaseDatabase.instance.ref("/users/${FirebaseAuth.instance.currentUser!.uid}");
  final snapshotUser = await ref.get();
  if (snapshotUser.exists) {
    userExists = true;
  }

  //If user does not exist return
  if (!userExists) {
    log("User does not exist");
    return 1;
  }

  //If yes: fetch users tickets
  ref = FirebaseDatabase.instance.ref("/users/${FirebaseAuth.instance.currentUser!.uid}/tickets");
  final snapshotTickets = await ref.get();
  if (snapshotTickets.exists) {
    tickets = snapshotTickets.value;
  }

  List<String> splitTickets = tickets.split(',');
  var ticketData;

  userTickets = [];
  //Retrieve tickets
  for (var t in splitTickets) {
    ref = FirebaseDatabase.instance.ref("tickets/$t");
    final snapshotTicket = await ref.get();
    if (snapshotTicket.exists) {
      final map = snapshotTicket.value as Map<Object?, Object?>;
      map.forEach((key, value) {
        ticketData = TicketData.fromJson(map);
      });
    }
    userTickets.add(ticketData);
  }

  return 0;
}


///Gets information about all tickets from firebase
Future<bool>firebaseGetAllTickets() async {

  Object? tickets;
  TicketData oneTicket = TicketData(activatedBy: "", activationDate: "", endDate: "", purchaseDate: "", purchasedBy: "", type: "", giftCode: "");

  DatabaseReference ref = FirebaseDatabase.instance.ref("/tickets");
  DataSnapshot ticketSnapshot = await ref.get();

  if (ticketSnapshot.exists) {
    tickets = ticketSnapshot.value;

    allUserTickets = [];

    final allTicketsMap = tickets as Map<Object?, dynamic>;
    allTicketsMap.forEach((key, value) {
      final oneTicketMap = value as Map<Object?, dynamic>;
      oneTicketMap.forEach((key, value) {
        oneTicket = TicketData.fromJson(oneTicketMap);
      });
      oneTicket.ticketID = key as String;
      allUserTickets.add(oneTicket);
      });
  }
  return true;
}

///Getting all VitenKode from the database
Future<bool>firebaseGetVitenKode() async {
  Object? tasks;
  TaskData oneTask = TaskData(task: "", one: "", two: "", three: "", four: "", five: "");

  DatabaseReference ref = FirebaseDatabase.instance.ref("/vitenkode");
  DataSnapshot userSnapshot = await ref.get();

  if (userSnapshot.exists) {
    tasks = userSnapshot.value;

    allTaskData = [];

    final allUsersMap = tasks as Map<Object?, dynamic>;
    allUsersMap.forEach((key, value) {
      final oneUserMap = value as Map<Object?, dynamic>;
      oneUserMap.forEach((key, value) {
        oneTask = TaskData.fromJson(oneUserMap);
      });
      oneTask.userID = key as String;
      allTaskData.add(oneTask);
    });
  }
  return true;
}

///Gets information about all users from firebase
Future<bool>firebaseGetAllUsers() async{

  Object? users;
  UserData oneUser = UserData(birthday: "", email: "", image: "", isAdmin: false, isEmployee: false, name: "", phone: "", tickets: "");

  DatabaseReference ref = FirebaseDatabase.instance.ref("/users");
  DataSnapshot userSnapshot = await ref.get();

  if (userSnapshot.exists) {
    users = userSnapshot.value;

    allUserData = [];

    final allUsersMap = users as Map<Object?, dynamic>;
    allUsersMap.forEach((key, value) {
      final oneUserMap = value as Map<Object?, dynamic>;
      oneUserMap.forEach((key, value) {
        oneUser = UserData.fromJson(oneUserMap);
      });
      oneUser.userID = key as String;
      allUserData.add(oneUser);
    });
  }
  return true;
}

///Update user data in the database (name, phone number, email, birthday, and/or admin/employee privileges)
///using the [user] object
Future <bool> firebaseUpdateUserInfo(UserData user) async {

  if (user.userID != "") {
    FirebaseDatabase.instance.ref("/users/${user.userID}").update(
      user.toJson()
    );
  }

  return true;
}

///Add [ticket] to database, and to a [user] if a [userID] is provided
///Return true for no issues, false for issues met
Future<bool> firebaseAddTicketToDatabase(TicketData ticket, UserData user, [String? userID]) async {

  String? ticketID;

  //Have to be authenticated to be able to add a ticket
  if (FirebaseAuth.instance.currentUser != null) {

    DatabaseReference ref = FirebaseDatabase.instance.ref("/tickets").push();
    ticketID = ref.key;

    await ref.set(ticket.toJson());
  }
  else {
    return false;
  }

  allUserTickets.add(ticket);

  //Only add ticket to a user if userID is provided
  if (ticketID != null && userID != "") {

    if (user.tickets == "") {
      user.tickets = "${user.tickets}$ticketID";
    } else {
      user.tickets = "${user.tickets},$ticketID";
    }

    var updates = {
      "tickets": user.tickets
    };

    FirebaseDatabase.instance.ref("/users/$userID").update(
      updates
    );
  }

   return true;
}

/// Adds several [tickets] to the database for a [user] using [userID]
Future<bool> firebaseAddMultipleTicketsToDatabase(List <TicketData> tickets, UserData user, String? userID) async{
  for (var ticket in tickets){
    if (!await firebaseAddTicketToDatabase(ticket, user, userID)){
      return false;
    }
  }
  return true;
}


///Renew an expired yearly membership [ticket]
Future <void> firebaseUpdateTicket(TicketData ticket) async {

  if (ticket.ticketID != "") {
    FirebaseDatabase.instance.ref("/tickets/${ticket.ticketID}").update(
        ticket.toJson()
    );
  }

}

/// Use a [seed] to figure out which default profile picture a user should get, then create a user using [name],
/// [mail], [phone] and [birthday]
Future<String> firebaseAddNonAuthenticatedUser(int seed, String name, String mail, String phone, String birthday) async {

  // Using math to figure out which profile picture to use
  // This is mostly yo create som variety
  int whichNum;
  if (seed % 10 == 0) {
    whichNum = 1;
  } else if (seed % 2 == 0) {
    whichNum = 2;
  } else {
    whichNum = 3;
  }

  // Get link to profile picture
  var downloadURL = await FirebaseStorage.instance.ref("/userPhotos/avatar$whichNum.png").getDownloadURL();

  // Define a user object
  var user = {
    "name": name,
    "email": mail,
    "phone": phone,
    "birthday": birthday,
    "isAdmin": false,
    "isEmployee": false,
    "image": downloadURL,
    "tickets": ""
  };

  // Push user object to database
  var ref = FirebaseDatabase.instance.ref("/users").push();
  var key = ref.key;

  ref.set(user);

  if (key != null) {
    return key;
  } else {
    return "";
  }
}

///Get user data from firebase realtime database
///Returns 0 for success, 1 for unable to get user data
Future<int> firebaseGetUser() async {

  //Location of data in database
  DatabaseReference ref = FirebaseDatabase.instance.ref("/users/${FirebaseAuth.instance.currentUser!.uid}");

  //Get data
  final snapshotUser = await ref.get();

  //If data exists iterate over it
  if(snapshotUser.exists) {
    final map = snapshotUser.value as Map<Object?, dynamic>;
    map.forEach((key, value) {
      //Current user holds all data relevant to the logged in user
      currentUser = UserData.fromJson(map);
      currentUser.userID = FirebaseAuth.instance.currentUser!.uid;
    });
  } else {
    return 1;
  }

  return 0;
}

///Update user [profilePic]
///Return 0 for success, 1 for no user, 2 for error during upload
Future<int>firebaseUpdateProfilePicture(File profilePic) async{

  //Get user id
  var uid = FirebaseAuth.instance.currentUser!.uid;

  //Ensure that someone is logged in
  if (uid.isNotEmpty) {

    //Location of image to be updated
    final storageRef = FirebaseStorage.instance.ref("userPhotos/$uid");

    String profilePicString = "";

    //Try updating the file and get the link to the file
    try {
      await storageRef.putFile(profilePic);
      profilePicString = await storageRef.getDownloadURL();
    } catch (e) {
      log("Error uploading new profile image, Error $e");
      return 2;
    }

    //Updates to be sent to the users info in Firebase
    var updates = {
      'image': profilePicString
    };

    //Update user info in Firebase
    FirebaseDatabase.instance.ref().child("/users/$uid").update(
      updates
    );

    //Update authentication data
    FirebaseAuth.instance.currentUser!.updatePhotoURL(profilePicString);

  } else {
    return 1;
  }

  return 0;
}

/// Update user [newMail] in all relevant fields, [password] needed for re-authentication
/// Returns 0 for success, 1 for error logging in, 2 for wrong password,
/// 3 for error updating authenticate credentials, 4 for error updating database
Future<int> firebaseUpdateMail(String newMail, String password) async{
  //Re-authenticate user
  //This is required by firebase before updating user authentication data
  try {
    final credential = await FirebaseAuth.instance.signInWithEmailAndPassword(
        email: currentUser.email,
        password: password
    );
    if (credential.credential != null) {
      await FirebaseAuth.instance.currentUser?.reauthenticateWithCredential(credential.credential!);
    }

  } on FirebaseAuthException catch (e) {
    if (e.code == 'user-not-found') {
      log('No user found for that email.');
      return 2;
    } else if (e.code == 'wrong-password') {
      log('Wrong password provided for that user.');
      return 2;
    }
  }

  //Update user authentication data
  try {
    await FirebaseAuth.instance.currentUser!.updateEmail(newMail);
  } catch (e) {
    log("Could not update user authentication, Error $e");
    return 3;
  }

  //Updates to be made
  var updates = {
    'email': newMail
  };

  //Update users data in Firebase
  try {
    FirebaseDatabase.instance.ref().child("/users/${FirebaseAuth.instance.currentUser?.uid}").update(
        updates
    );
  } catch (e) {
    log("Could not update database, Error $e");
    return 4;
  }

  //Update local variable
  currentUser.email = newMail;

  return 0;
}

/// Update [oldPassword] to be [newPassword].
/// Return 0 for success, 2 for wrong password, 3 for problems updating password
Future<int> firebaseUpdatePassword(String oldPassword, String newPassword) async{

  //Re-authenticate user, by first signing them in again
  //This is required by Firebase before user authentication data is changed
  try {
    final credential = await FirebaseAuth.instance.signInWithEmailAndPassword(
        email: currentUser.email,
        password: oldPassword
    );
    if (credential.credential != null) {
      await FirebaseAuth.instance.currentUser?.reauthenticateWithCredential(credential.credential!);
    }

  } on FirebaseAuthException catch (e) {
    if (e.code == 'user-not-found') {
      log('No user found for that email.');
      return 2;
    } else if (e.code == 'wrong-password') {
      log('Wrong password provided for that user.');
      return 2;
    }
  }

  //Update password in firebase authentication
  try {
    await FirebaseAuth.instance.currentUser!.updatePassword(newPassword);
  } catch (e) {
    log("Could not update user authentication, Error $e");
    return 3;
  }

  return 0;
}

///Update [phone] number in database,
///Returns 0 for success, 1 for errors during updating
Future<int> firebaseUpdatePhone(String phone) async{

  //List of updates
  var updates = {
    'phone': phone
  };

  //Update current users data with what is stored in updates
  try {
    FirebaseDatabase.instance.ref().child("/users/${FirebaseAuth.instance.currentUser?.uid}").update(
        updates
    );
  } catch (e) {
    log("Could not update database, Error $e");
    return 1;
  }

  //Update local variable
  currentUser.phone = phone;


  return 0;
}

/// Delete user from database using [userID]. If set to true will also [deletImage] and [deleteAuth]
/// Returns 0 for success, 1 for could not delete profile pic,
/// 2 for could not delete user data, 3 for could not delete authentication
Future <int> firebaseDeleteUser(String userID, bool deleteImage, bool deleteAuth) async {

  //Delete profile picture
  if (deleteImage) {
    final ref = FirebaseStorage.instance.ref("/userPhotos/$userID");
    try {
      await ref.delete();
    } catch (e) {
      log("Could no delete profile image, Error: $e");
      return 1;
    }
  }

  //Delete user data saved in Firebase Realtime Database
  final ref = FirebaseDatabase.instance.ref("/users/$userID");
  try {
    await ref.remove();
  } catch (e) {
    log("Could no delete user data, Error: $e");
    return 2;
  }

  //Delete the user from Firebase Authentication
  if (deleteAuth) {
    try {
      await FirebaseAuth.instance.currentUser?.delete();
    } catch (e) {
      log("Could no delete Firebase Authentication, Error: $e");
      return 3;
    }
  }

  return 0;
}

///Get last 100 blog entries from database, return 0 for no issues, 1 for no entries found
Future<int> firebaseGetBlog() async {
  var posts;
  BlogData postData = BlogData(content: "", image: "", notify: false, publicationDate: "", title: "", author: "");
  if (allBlogs.isNotEmpty) {
    allBlogs = [];
  }

  //Get data from database
  DatabaseReference ref = FirebaseDatabase.instance.ref("/posts");
  final postRef = FirebaseStorage.instance.ref();
  DataSnapshot  postSnapshot = await ref.limitToLast(100).get();

  //Check if data were retrieved
  if (postSnapshot.exists) {
    posts = postSnapshot.value;
  } else {
    return 1;
  }

  //Sort data into list
  final allBlogsMap = posts as Map<Object?, Object?>;
  allBlogsMap.forEach((key, value) {
    //print(value);
    final oneBlogMap = value as Map<Object?, Object?>;
    oneBlogMap.forEach((key, value) {
      postData = BlogData.fromJson(oneBlogMap);
    });
    postData.blogID = key as String;
    allBlogs.add(postData);
  });

  //Get actual blog content
  for (var element in allBlogs) {
    var onePostRef = postRef.child(element.content);

    try {
      final Uint8List? data = await onePostRef.getData(1024*1024);
      String s = String.fromCharCodes(data!);
      element.postContent = s;
    } on FirebaseException catch (e) {
      log("Issue getting blog content, error: $e");
    }

  }
  return 0;
}

///Upload [image] to the storage under [uniqueFileName]
Future <bool> firebaseUploadImageForBlog(Uint8List image, String uniqueFileName) async {

  try{
    Reference referenceRoot = FirebaseStorage.instance.ref();
    Reference referenceDirImages = referenceRoot.child('/blogPhotos/$uniqueFileName');
    await referenceDirImages.putData(image);
  } catch(error){
    return false;
  }
  return true;
}

///Function uploads a blog post, [tempBlogPost], to the database, with its image
Future<bool> firebaseUploadBlog(BlogData tempBlogPost) async {

  String? blogID;

  //Have to be authenticated to be able to add the post
  if (FirebaseAuth.instance.currentUser != null) {

    DatabaseReference ref = FirebaseDatabase.instance.ref("/posts").push();
    blogID = ref.key;

    await ref.set(tempBlogPost.toJson());
  }
  else {
    return false;
  }
  allBlogs.add(tempBlogPost);

  var updates = {
    'content':  tempBlogPost.content,
    'image': tempBlogPost.image,
    'notify': tempBlogPost.notify,
    'publicationDate': tempBlogPost.publicationDate,
    'title': tempBlogPost.title,
    'author': tempBlogPost.author,
  };

  FirebaseDatabase.instance.ref("/posts/$blogID").update(
      updates
  );
  return false;
}



///Update [blog] data
Future <bool> firebaseUpdateBlog(BlogData blog) async {

  var updates = {
    'content':  blog.content,
    'image': blog.image,
    'notify': blog.notify,
    'publicationDate': blog.publicationDate,
    'title': blog.title,
    'author': blog.author,
  };

  if (blog.title != "") {
    FirebaseDatabase.instance.ref("/posts/${blog.blogID}").update(
        updates
    );
  }

  return true;
}

///Delete [blog] data
Future <bool> firebaseDeleteBlog(BlogData blog) async {
  final ref = FirebaseDatabase.instance.ref();
  ref.child('posts/${blog.blogID}').remove();
  return true;
}

